@@ -17,7 +17,7 @@
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  * 
- * $Id: dbi_result.c,v 1.45 2006/06/02 17:17:56 qu1j0t3 Exp $
+ * $Id: dbi_result.c,v 1.57 2013/01/08 23:54:30 mhoenicka Exp $
  *
  * (anything that has to do with row seeking or fetching fields goes in this file)
  */
@@ -30,14 +30,18 @@
 #include <stdlib.h>
 #include <stdarg.h>
 #include <string.h>
-#include <unistd.h>
+#ifdef HAVE_UNISTD_H
+  #include <unistd.h>
+#endif
 #include <math.h>
 #include <limits.h>
+#include <time.h>
 
 #include <dbi/dbi.h>
 #include <dbi/dbi-dev.h>
 
 #ifdef __MINGW32__
+#warning "Using the thread-unsafe strtok function!"
 #define strtok_r(s1,s2,s3) strtok(s1,s2)
 #endif
 
@@ -83,6 +87,8 @@ int dbi_result_seek_row(dbi_result Result, unsigned long long rowidx) {
     return 0;
   }	
 
+  _reset_conn_error(RESULT->conn);
+
   if (RESULT->result_state == NOTHING_RETURNED 
       || rowidx <= 0 || rowidx > RESULT->numrows_matched) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
@@ -97,7 +103,7 @@ int dbi_result_seek_row(dbi_result Result, unsigned long long rowidx) {
   }
 	
   /* row is one-based for the user, but zero-based to the dbd conn */
-  retval = RESULT->conn->driver->functions->goto_row(RESULT, rowidx-1);
+  retval = RESULT->conn->driver->functions->goto_row(RESULT, rowidx-1, RESULT->currowidx-1);
   if (retval == -1) {
     _error_handler(RESULT->conn, DBI_ERROR_DBD);
     return 0;
@@ -127,6 +133,8 @@ int dbi_result_has_prev_row(dbi_result Result) {
     return 0;
   }	
 
+  _reset_conn_error(RESULT->conn);
+
   return RESULT->result_state != NOTHING_RETURNED && RESULT->currowidx > 1;
 }
 
@@ -136,6 +144,8 @@ int dbi_result_prev_row(dbi_result Result) {
     return 0;
   }	
 
+  _reset_conn_error(RESULT->conn);
+
   if (!dbi_result_has_prev_row(Result)) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
     return 0;
@@ -149,6 +159,8 @@ int dbi_result_has_next_row(dbi_result Result) {
     return 0;
   }	
 
+  _reset_conn_error(RESULT->conn);
+
   return RESULT->result_state != NOTHING_RETURNED
   		 && RESULT->currowidx < dbi_result_get_numrows(Result);
 }
@@ -159,6 +171,8 @@ int dbi_result_next_row(dbi_result Result) {
     return 0;
   }	
 
+  _reset_conn_error(RESULT->conn);
+
   if (!dbi_result_has_next_row(Result)) {
 /*     _error_handler(RESULT->conn, DBI_ERROR_BADIDX); */
     return 0;
@@ -172,6 +186,8 @@ unsigned long long dbi_result_get_currow(dbi_result Result) {
     return 0;
   }	
 
+  _reset_conn_error(RESULT->conn);
+
   return RESULT->currowidx;
 }
 
@@ -181,6 +197,8 @@ unsigned long long dbi_result_get_numrows(dbi_result Result) {
     return DBI_ROW_ERROR;
   }	
 
+  _reset_conn_error(RESULT->conn);
+
   return RESULT->numrows_matched;
 }
 
@@ -190,19 +208,23 @@ unsigned long long dbi_result_get_numrows_affected(dbi_result Result) {
     return DBI_ROW_ERROR;
   }	
 
+  _reset_conn_error(RESULT->conn);
+
   return RESULT->numrows_affected;
 }
 
 /* returns the length of the string or binary, excluding a trailing \0 */
 size_t dbi_result_get_field_length(dbi_result Result, const char *fieldname) {
   unsigned int fieldidx = 0;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
 
   if (!RESULT) {
     _error_handler(/*RESULT->conn*/ NULL, DBI_ERROR_BADPTR);
     return DBI_LENGTH_ERROR;
   }
 
+  _reset_conn_error(RESULT->conn);
+
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     _error_handler(RESULT->conn, errflag);
@@ -223,6 +245,8 @@ size_t dbi_result_get_field_length_idx(dbi_result Result, unsigned int fieldidx)
     return DBI_LENGTH_ERROR;
   }
 
+  _reset_conn_error(RESULT->conn);
+
   currowidx = RESULT->currowidx;
   if (!RESULT->rows[currowidx] || !RESULT->rows[currowidx]->field_sizes) {
     _error_handler(RESULT->conn, DBI_ERROR_BADOBJECT);
@@ -251,13 +275,15 @@ size_t dbi_result_get_field_size_idx(dbi_result Result, unsigned int fieldidx) {
 
 unsigned int dbi_result_get_field_idx(dbi_result Result, const char *fieldname) {
   unsigned int fieldidx = 0;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
 
   if (!RESULT) {
     _error_handler(/*RESULT->conn*/ NULL, DBI_ERROR_BADPTR);
     return 0;
   }	
 
+  _reset_conn_error(RESULT->conn);
+
   /* user-visible indexes start at 1 */
   fieldidx = _find_field(RESULT, fieldname, &errflag)+1;
   if (errflag != DBI_ERROR_NONE) {
@@ -275,7 +301,9 @@ const char *dbi_result_get_field_name(dbi_result Result, unsigned int fieldidx)
     return NULL;
   }
 
-  else if (fieldidx > RESULT->numfields) {
+  _reset_conn_error(RESULT->conn);
+
+  if (fieldidx > RESULT->numfields) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
     return NULL;
   }
@@ -289,18 +317,23 @@ const char *dbi_result_get_field_name(dbi_result Result, unsigned int fieldidx)
 
 unsigned int dbi_result_get_numfields(dbi_result Result) {
   if (!RESULT) return DBI_FIELD_ERROR;
+
+  _reset_conn_error(RESULT->conn);
+
   return RESULT->numfields;
 }
 
 unsigned short dbi_result_get_field_type(dbi_result Result, const char *fieldname) {
   unsigned int fieldidx = 0;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
 	
   if (!RESULT) {
     _error_handler(/*RESULT->conn*/ NULL, DBI_ERROR_BADPTR);
     return DBI_TYPE_ERROR;
   }
 
+  _reset_conn_error(RESULT->conn);
+
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     _error_handler(RESULT->conn, errflag);
@@ -318,7 +351,9 @@ unsigned short dbi_result_get_field_type_idx(dbi_result Result, unsigned int fie
     return DBI_TYPE_ERROR;
   }
 
-  else if (!RESULT->field_types) {
+  _reset_conn_error(RESULT->conn);
+
+  if (!RESULT->field_types) {
     _error_handler(RESULT->conn, DBI_ERROR_BADOBJECT);
     return DBI_TYPE_ERROR;
   }
@@ -332,13 +367,15 @@ unsigned short dbi_result_get_field_type_idx(dbi_result Result, unsigned int fie
 
 unsigned int dbi_result_get_field_attrib(dbi_result Result, const char *fieldname, unsigned int attribmin, unsigned int attribmax) {
   unsigned int fieldidx = 0;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
 	
   if (!RESULT) {
     _error_handler(/*RESULT->conn*/ NULL, DBI_ERROR_BADPTR);
     return DBI_ATTRIBUTE_ERROR;
   }
 
+  _reset_conn_error(RESULT->conn);
+
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     _error_handler(RESULT->conn, errflag);
@@ -356,7 +393,10 @@ unsigned int dbi_result_get_field_attrib_idx(dbi_result Result,
     _error_handler(/*RESULT->conn*/ NULL, DBI_ERROR_BADPTR);
     return DBI_ATTRIBUTE_ERROR;
   }
-  else if (!RESULT->field_attribs) {
+
+  _reset_conn_error(RESULT->conn);
+
+  if (!RESULT->field_attribs) {
     _error_handler(RESULT->conn, DBI_ERROR_BADOBJECT);
     return DBI_ATTRIBUTE_ERROR;
   }
@@ -370,13 +410,15 @@ unsigned int dbi_result_get_field_attrib_idx(dbi_result Result,
 
 unsigned int dbi_result_get_field_attribs(dbi_result Result, const char *fieldname) {
   unsigned int fieldidx = 0;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
 	
   if (!RESULT) {
     _error_handler(/*RESULT->conn*/ NULL, DBI_ERROR_BADPTR);
     return DBI_ATTRIBUTE_ERROR;
   }
 
+  _reset_conn_error(RESULT->conn);
+
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     _error_handler(RESULT->conn, errflag);
@@ -393,7 +435,10 @@ unsigned int dbi_result_get_field_attribs_idx(dbi_result Result, unsigned int fi
     _error_handler(/*RESULT->conn*/ NULL, DBI_ERROR_BADPTR);
     return DBI_ATTRIBUTE_ERROR;
   }
-  else if (!RESULT->field_attribs) {
+
+  _reset_conn_error(RESULT->conn);
+
+  if (!RESULT->field_attribs) {
     _error_handler(RESULT->conn, DBI_ERROR_BADOBJECT);
     return DBI_ATTRIBUTE_ERROR;
   }
@@ -419,13 +464,15 @@ int _get_field_flag(dbi_row_t *row, unsigned int fieldidx, unsigned char flag) {
 
 int dbi_result_field_is_null(dbi_result Result, const char *fieldname) {
   unsigned int fieldidx = 0;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
 
   if (!RESULT) {
     _error_handler(/*RESULT->conn*/ NULL, DBI_ERROR_BADPTR);
     return DBI_FIELD_FLAG_ERROR;
   }
 
+  _reset_conn_error(RESULT->conn);
+
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     _error_handler(RESULT->conn, errflag);
@@ -445,6 +492,8 @@ int dbi_result_field_is_null_idx(dbi_result Result, unsigned int fieldidx) {
     return DBI_FIELD_FLAG_ERROR;
   }
 
+  _reset_conn_error(RESULT->conn);
+
   currowidx = RESULT->currowidx;
   if (!RESULT->rows[currowidx] || !RESULT->rows[currowidx]->field_flags) {
     _error_handler(RESULT->conn, DBI_ERROR_BADOBJECT);
@@ -502,6 +551,8 @@ int dbi_result_free(dbi_result Result) {
     retval = _disjoin_from_conn(RESULT);
   }
 
+  _reset_conn_error(RESULT->conn);
+
   while (RESULT->field_bindings) {
     _remove_binding_node(RESULT, RESULT->field_bindings);
   }
@@ -761,35 +812,37 @@ unsigned int dbi_result_bind_fields(dbi_result Result, const char *format, ...)
 /* RESULT: get_* functions */
 
 signed char dbi_result_get_char(dbi_result Result, const char *fieldname) {
-  signed char ERROR = 0;
+  signed char my_ERROR = 0;
   unsigned int fieldidx;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
+
+  _reset_conn_error(RESULT->conn);
 
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     dbi_conn_t *conn = RESULT->conn;
     _error_handler(conn, errflag);
-    return ERROR;
+    return my_ERROR;
   }
   return dbi_result_get_char_idx(Result, fieldidx+1);
 }
 
 signed char dbi_result_get_char_idx(dbi_result Result, unsigned int fieldidx) {
-  signed char ERROR = 0;
+  signed char my_ERROR = 0;
 
   fieldidx--;
 
-  RESULT->conn->error_flag = DBI_ERROR_NONE;
+  _reset_conn_error(RESULT->conn);
 
   if (fieldidx >= RESULT->numfields) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
-    return ERROR;
+    return my_ERROR;
   }
   if (RESULT->field_types[fieldidx] != DBI_TYPE_INTEGER) {
-    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
     _verbose_handler(RESULT->conn, "%s: field `%s` is not integer type\n",
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
-    return ERROR;
+    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+    return my_ERROR;
   }
 
   switch (RESULT->field_attribs[fieldidx] & DBI_INTEGER_SIZEMASK) {
@@ -803,43 +856,46 @@ signed char dbi_result_get_char_idx(dbi_result Result, unsigned int fieldidx) {
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
   default:
     _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
-    return ERROR;
+    return my_ERROR;
   }
 }
 
 short dbi_result_get_short(dbi_result Result, const char *fieldname) {
-  short ERROR = 0;
+  short my_ERROR = 0;
   unsigned int fieldidx;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
+
+  _reset_conn_error(RESULT->conn);
 
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     dbi_conn_t *conn = RESULT->conn;
     _error_handler(conn, DBI_ERROR_BADNAME);
-    return ERROR;
+    return my_ERROR;
   }
   return dbi_result_get_short_idx(Result, fieldidx+1);
 }
 
 short dbi_result_get_short_idx(dbi_result Result, unsigned int fieldidx) {
-  short ERROR = 0;
+  short my_ERROR = 0;
   fieldidx--;
 
-  RESULT->conn->error_flag = DBI_ERROR_NONE;
+  _reset_conn_error(RESULT->conn);
 
   if (fieldidx >= RESULT->numfields) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
-    return ERROR;
+    return my_ERROR;
   }
   if (RESULT->field_types[fieldidx] != DBI_TYPE_INTEGER) {
-    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
     _verbose_handler(RESULT->conn, "%s: field `%s` is not integer type\n",
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
-    return ERROR;
+    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+    return my_ERROR;
   }
 
   switch (RESULT->field_attribs[fieldidx] & DBI_INTEGER_SIZEMASK) {
   case DBI_INTEGER_SIZE1:
+    return RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_char;
   case DBI_INTEGER_SIZE2:
     return RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_short;
   case DBI_INTEGER_SIZE3:
@@ -849,7 +905,7 @@ short dbi_result_get_short_idx(dbi_result Result, unsigned int fieldidx) {
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
   default:
     _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
-    return ERROR;
+    return my_ERROR;
   }
 }
 
@@ -858,17 +914,19 @@ int dbi_result_get_long(dbi_result Result, const char *fieldname) {
 }
 
 int dbi_result_get_int(dbi_result Result, const char *fieldname) {
-  long ERROR = 0;
+  long my_ERROR = 0;
   unsigned int fieldidx;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
+
+  _reset_conn_error(RESULT->conn);
 
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     dbi_conn_t *conn = RESULT->conn;
     _error_handler(conn, DBI_ERROR_BADNAME);
-    return ERROR;
+    return my_ERROR;
   }
-  return dbi_result_get_long_idx(Result, fieldidx+1);
+  return dbi_result_get_int_idx(Result, fieldidx+1);
 }
 
 int dbi_result_get_long_idx(dbi_result Result, unsigned int fieldidx) {
@@ -876,25 +934,27 @@ int dbi_result_get_long_idx(dbi_result Result, unsigned int fieldidx) {
 }
 
 int dbi_result_get_int_idx(dbi_result Result, unsigned int fieldidx) {
-  long ERROR = 0;
+  long my_ERROR = 0;
   fieldidx--;
 	
-  RESULT->conn->error_flag = DBI_ERROR_NONE;
+  _reset_conn_error(RESULT->conn);
 
   if (fieldidx >= RESULT->numfields) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
-    return ERROR;
+    return my_ERROR;
   }
   if (RESULT->field_types[fieldidx] != DBI_TYPE_INTEGER) {
-    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
     _verbose_handler(RESULT->conn, "%s: field `%s` is not integer type\n",
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
-    return ERROR;
+    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+    return my_ERROR;
   }
 
   switch (RESULT->field_attribs[fieldidx] & DBI_INTEGER_SIZEMASK) {
   case DBI_INTEGER_SIZE1:
+    return RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_char;
   case DBI_INTEGER_SIZE2:
+    return RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_short;
   case DBI_INTEGER_SIZE3:
   case DBI_INTEGER_SIZE4:
     return RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_long;
@@ -903,51 +963,56 @@ int dbi_result_get_int_idx(dbi_result Result, unsigned int fieldidx) {
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
   default:
     _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
-    return ERROR;
+    return my_ERROR;
   }
 }
 
 long long dbi_result_get_longlong(dbi_result Result, const char *fieldname) {
-  long long ERROR = 0;
+  long long my_ERROR = 0;
   unsigned int fieldidx;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
     
+  _reset_conn_error(RESULT->conn);
+
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     dbi_conn_t *conn = RESULT->conn;
     _error_handler(conn, DBI_ERROR_BADNAME);
-    return ERROR;
+    return my_ERROR;
   }
   return dbi_result_get_longlong_idx(Result, fieldidx+1);
 }
 
 long long dbi_result_get_longlong_idx(dbi_result Result, unsigned int fieldidx) {
-  long long ERROR = 0;
+  long long my_ERROR = 0;
   fieldidx--;
 	
-  RESULT->conn->error_flag = DBI_ERROR_NONE;
+  _reset_conn_error(RESULT->conn);
 
   if (fieldidx >= RESULT->numfields) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
-    return ERROR;
+    return my_ERROR;
   }
   if (RESULT->field_types[fieldidx] != DBI_TYPE_INTEGER) {
-    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
     _verbose_handler(RESULT->conn, "%s: field `%s` is not integer type\n",
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
-    return ERROR;
+    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+    return my_ERROR;
   }
 
   switch (RESULT->field_attribs[fieldidx] & DBI_INTEGER_SIZEMASK) {
   case DBI_INTEGER_SIZE1:
+    return RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_char;
   case DBI_INTEGER_SIZE2:
+    return RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_short;
   case DBI_INTEGER_SIZE3:
   case DBI_INTEGER_SIZE4:
+    return RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_long;
   case DBI_INTEGER_SIZE8:
     return RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_longlong;
   default:
     _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
-    return ERROR;
+    return my_ERROR;
   }
 }
 
@@ -992,34 +1057,37 @@ unsigned long long dbi_result_get_ulonglong_idx(dbi_result Result, unsigned int
 }
 
 float dbi_result_get_float(dbi_result Result, const char *fieldname) {
-  float ERROR = 0.0;
-  dbi_error_flag errflag;
+  float my_ERROR = 0.0;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
   unsigned int fieldidx;
 
+  _reset_conn_error(RESULT->conn);
+
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     dbi_conn_t *conn = RESULT->conn;
     _error_handler(conn, DBI_ERROR_BADNAME);
-    return ERROR;
+    return my_ERROR;
   }
   return dbi_result_get_float_idx(Result, fieldidx+1);
 }
 
 float dbi_result_get_float_idx(dbi_result Result, unsigned int fieldidx) {
-  float ERROR = 0.0;
+  float my_ERROR = 0.0;
   fieldidx--;
 	
-  RESULT->conn->error_flag = DBI_ERROR_NONE;
-	
+  _reset_conn_error(RESULT->conn);
+
+
   if (fieldidx >= RESULT->numfields) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
-    return ERROR;
+    return my_ERROR;
   }
   if (RESULT->field_types[fieldidx] != DBI_TYPE_DECIMAL) {
-    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
     _verbose_handler(RESULT->conn, "%s: field `%s` is not float type\n",
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
-    return ERROR;
+    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+    return my_ERROR;
   }
 
   switch (RESULT->field_attribs[fieldidx] & DBI_DECIMAL_SIZEMASK) {
@@ -1030,80 +1098,88 @@ float dbi_result_get_float_idx(dbi_result Result, unsigned int fieldidx) {
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
   default:
     _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
-    return ERROR;
+    return my_ERROR;
   }
 }
 
 double dbi_result_get_double(dbi_result Result, const char *fieldname) {
-  double ERROR = 0.0;
+  double my_ERROR = 0.0;
   unsigned int fieldidx;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
+
+  _reset_conn_error(RESULT->conn);
 
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     dbi_conn_t *conn = RESULT->conn;
     _error_handler(conn, DBI_ERROR_BADNAME);
-    return ERROR;
+    return my_ERROR;
   }
   return dbi_result_get_double_idx(Result, fieldidx+1);
 }
 	
 double dbi_result_get_double_idx(dbi_result Result, unsigned int fieldidx) {
-  double ERROR = 0.0;
+  double my_ERROR = 0.0;
   fieldidx--;
 	
-  RESULT->conn->error_flag = DBI_ERROR_NONE;
-	
+  _reset_conn_error(RESULT->conn);
+
   if (fieldidx >= RESULT->numfields) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
-    return ERROR;
+    return my_ERROR;
   }
   if (RESULT->field_types[fieldidx] != DBI_TYPE_DECIMAL) {
-    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
     _verbose_handler(RESULT->conn, "%s: field `%s` is not double type\n",
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
-    return ERROR;
+    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+    return my_ERROR;
   }
 
   switch (RESULT->field_attribs[fieldidx] & DBI_DECIMAL_SIZEMASK) {
   case DBI_DECIMAL_SIZE4:
+    return RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_float;
   case DBI_DECIMAL_SIZE8:
     return RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_double;
   default:
     _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
-    return ERROR;
+    return my_ERROR;
   }
 }
 
 const char *dbi_result_get_string(dbi_result Result, const char *fieldname) {
-  const char *ERROR = "ERROR";
+  const char *my_ERROR = "ERROR";
   unsigned int fieldidx;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
+
+  _reset_conn_error(RESULT->conn);
 
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     dbi_conn_t *conn = RESULT->conn;
     _error_handler(conn, DBI_ERROR_BADNAME);
-    return ERROR;
+    return my_ERROR;
   }
   return dbi_result_get_string_idx(Result, fieldidx+1);
 }
 	
 const char *dbi_result_get_string_idx(dbi_result Result, unsigned int fieldidx) {
-  const char *ERROR = "ERROR";
+  const char *my_ERROR = "ERROR";
   fieldidx--;
 	
-  RESULT->conn->error_flag = DBI_ERROR_NONE;
-	
+  _reset_conn_error(RESULT->conn);
+
+
   if (fieldidx >= RESULT->numfields) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
-    return ERROR;
+    return my_ERROR;
   }
+
   if (RESULT->field_types[fieldidx] != DBI_TYPE_STRING) {
-    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
-    _verbose_handler(RESULT->conn, "%s: field `%s` is not string type\n",
+    dbi_conn_t *conn = RESULT->conn;
+    _verbose_handler(conn, "%s: field `%s` is not string type\n",
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
-    return ERROR;
+    _error_handler(conn, DBI_ERROR_BADTYPE);
+    return my_ERROR;
   }
   if (RESULT->rows[RESULT->currowidx]->field_sizes[fieldidx] == 0
       && _get_field_flag(RESULT->rows[RESULT->currowidx], fieldidx, DBI_VALUE_NULL)) {
@@ -1116,34 +1192,37 @@ const char *dbi_result_get_string_idx(dbi_result Result, unsigned int fieldidx)
 }
 
 const unsigned char *dbi_result_get_binary(dbi_result Result, const char *fieldname) {
-  const char *ERROR = "ERROR";
+  const char *my_ERROR = "ERROR";
   unsigned int fieldidx;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
+
+  _reset_conn_error(RESULT->conn);
 
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     dbi_conn_t *conn = RESULT->conn;
     _error_handler(conn, DBI_ERROR_BADNAME);
-    return (const unsigned char*)ERROR;
+    return (const unsigned char*)my_ERROR;
   }
   return dbi_result_get_binary_idx(Result, fieldidx+1);
 }
 	
 const unsigned char *dbi_result_get_binary_idx(dbi_result Result, unsigned int fieldidx) {
-  const char *ERROR = "ERROR";
+  const char *my_ERROR = "ERROR";
   fieldidx--;
 	
-  RESULT->conn->error_flag = DBI_ERROR_NONE;
-	
+  _reset_conn_error(RESULT->conn);
+
+
   if (fieldidx >= RESULT->numfields) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
-    return (const unsigned char*)ERROR;
+    return (const unsigned char*)my_ERROR;
   }
   if (RESULT->field_types[fieldidx] != DBI_TYPE_BINARY) {
-    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
     _verbose_handler(RESULT->conn, "%s: field `%s` is not binary type\n",
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
-    return (const unsigned char*)ERROR;
+    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+    return (const unsigned char*)my_ERROR;
   }
   if (RESULT->rows[RESULT->currowidx]->field_sizes[fieldidx] == 0) return NULL;
 
@@ -1151,35 +1230,38 @@ const unsigned char *dbi_result_get_binary_idx(dbi_result Result, unsigned int f
 }
 
 char *dbi_result_get_string_copy(dbi_result Result, const char *fieldname) {
-  char *ERROR = "ERROR";
+  char *my_ERROR = "ERROR";
   unsigned int fieldidx;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
+
+  _reset_conn_error(RESULT->conn);
 
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     dbi_conn_t *conn = RESULT->conn;
     _error_handler(conn, DBI_ERROR_BADNAME);
-    return strdup(ERROR);
+    return strdup(my_ERROR);
   }
   return dbi_result_get_string_copy_idx(Result, fieldidx+1);
 }
 	
 char *dbi_result_get_string_copy_idx(dbi_result Result, unsigned int fieldidx) {
-  char *ERROR = "ERROR";
+  char *my_ERROR = "ERROR";
   char *newstring = NULL;
   fieldidx--;
 	
-  RESULT->conn->error_flag = DBI_ERROR_NONE;
-	
+  _reset_conn_error(RESULT->conn);
+
+
   if (fieldidx >= RESULT->numfields) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
-    return strdup(ERROR);
+    return strdup(my_ERROR);
   }
   if (RESULT->field_types[fieldidx] != DBI_TYPE_STRING) {
-    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
     _verbose_handler(RESULT->conn, "%s: field `%s` is not string type\n",
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
-    return strdup(ERROR);
+    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+    return strdup(my_ERROR);
   }
   if (RESULT->rows[RESULT->currowidx]->field_sizes[fieldidx] == 0
    && RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_string == NULL) {
@@ -1195,88 +1277,280 @@ char *dbi_result_get_string_copy_idx(dbi_result Result, unsigned int fieldidx) {
   }
   else {
     _error_handler(RESULT->conn, DBI_ERROR_NOMEM);
-    return strdup(ERROR);
+    return strdup(my_ERROR);
   }
 }
 
 unsigned char *dbi_result_get_binary_copy(dbi_result Result, const char *fieldname) {
-  char *ERROR = "ERROR";
+  char *my_ERROR = "ERROR";
   unsigned int fieldidx;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
+
+  _reset_conn_error(RESULT->conn);
 
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     dbi_conn_t *conn = RESULT->conn;
     _error_handler(conn, DBI_ERROR_BADNAME);
-    return (unsigned char *)strdup(ERROR);
+    return (unsigned char *)strdup(my_ERROR);
   }
   return dbi_result_get_binary_copy_idx(Result, fieldidx+1);
 }
 	
 unsigned char *dbi_result_get_binary_copy_idx(dbi_result Result, unsigned int fieldidx) {
-  char *ERROR = "ERROR";
+  char *my_ERROR = "ERROR";
   unsigned char *newblob = NULL;
   unsigned long long size;
   fieldidx--;
 
-  RESULT->conn->error_flag = DBI_ERROR_NONE;
-	
+  _reset_conn_error(RESULT->conn);
+
+
   if (fieldidx >= RESULT->numfields) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
-    return (unsigned char *)strdup(ERROR);
+    return (unsigned char *)strdup(my_ERROR);
   }
   if (RESULT->field_types[fieldidx] != DBI_TYPE_BINARY) {
-    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
     _verbose_handler(RESULT->conn, "%s: field `%s` is not binary type\n",
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
-    return (unsigned char *)strdup(ERROR);
+    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+    return (unsigned char *)strdup(my_ERROR);
   }
   if (RESULT->rows[RESULT->currowidx]->field_sizes[fieldidx] == 0) return NULL;
 
-  size = dbi_result_get_field_size_idx(Result, fieldidx);
+  /* API function must use 1-based index */
+  size = dbi_result_get_field_length_idx(Result, fieldidx+1);
   newblob = malloc(size);
   if (!newblob) {
     _error_handler(RESULT->conn, DBI_ERROR_NOMEM);
-    return (unsigned char *)strdup(ERROR);
+    return (unsigned char *)strdup(my_ERROR);
   }
   memcpy(newblob, RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_string, size);
   return newblob;
 }
 
 time_t dbi_result_get_datetime(dbi_result Result, const char *fieldname) {
-  time_t ERROR = 0;
+  time_t my_ERROR = 0;
   unsigned int fieldidx;
-  dbi_error_flag errflag;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
+
+  _reset_conn_error(RESULT->conn);
 
   fieldidx = _find_field(RESULT, fieldname, &errflag);
   if (errflag != DBI_ERROR_NONE) {
     dbi_conn_t *conn = RESULT->conn;
     _error_handler(conn, DBI_ERROR_BADNAME);
-    return ERROR;
+    return my_ERROR;
   }
   return dbi_result_get_datetime_idx(Result, fieldidx+1);
 }
 	
 time_t dbi_result_get_datetime_idx(dbi_result Result, unsigned int fieldidx) {
-  time_t ERROR = 0;
+  time_t my_ERROR = 0;
   fieldidx--;
 
-  RESULT->conn->error_flag = DBI_ERROR_NONE;
-	
+  _reset_conn_error(RESULT->conn);
+
   if (fieldidx >= RESULT->numfields) {
     _error_handler(RESULT->conn, DBI_ERROR_BADIDX);
-    return ERROR;
+    return my_ERROR;
   }
   if (RESULT->field_types[fieldidx] != DBI_TYPE_DATETIME) {
-    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
     _verbose_handler(RESULT->conn, "%s: field `%s` is not datetime type\n",
                      __func__, dbi_result_get_field_name(Result, fieldidx+1));
-    return ERROR;
+    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+    return my_ERROR;
   }
 	
   return (time_t)(RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_datetime);
 }
 
+/* RESULT: get_as* functions */
+
+long long dbi_result_get_as_longlong(dbi_result Result, const char *fieldname) {
+  long long my_ERROR = 0;
+  unsigned int fieldidx;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
+    
+  _reset_conn_error(RESULT->conn);
+
+  fieldidx = _find_field(RESULT, fieldname, &errflag);
+  if (errflag != DBI_ERROR_NONE) {
+    dbi_conn_t *conn = RESULT->conn;
+    _error_handler(conn, DBI_ERROR_BADNAME);
+    return my_ERROR;
+  }
+  return dbi_result_get_as_longlong_idx(Result, fieldidx+1);
+}
+
+long long dbi_result_get_as_longlong_idx(dbi_result Result, unsigned int fieldidx) {
+  long long my_ERROR = 0;
+  fieldidx--;
+
+  switch (RESULT->field_types[fieldidx]) {
+  case DBI_TYPE_INTEGER:
+    switch (RESULT->field_attribs[fieldidx] & DBI_INTEGER_SIZEMASK) {
+    case DBI_INTEGER_SIZE1:
+      return (long long)RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_char;
+    case DBI_INTEGER_SIZE2:
+      return (long long)RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_short;
+    case DBI_INTEGER_SIZE3:
+    case DBI_INTEGER_SIZE4:
+      return (long long)RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_long;
+    case DBI_INTEGER_SIZE8:
+      return RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_longlong;
+    default:
+      _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+      return my_ERROR;
+    }
+  case DBI_TYPE_DECIMAL:
+    switch (RESULT->field_attribs[fieldidx] & DBI_DECIMAL_SIZEMASK) {
+    case DBI_DECIMAL_SIZE4:
+      return (long long)RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_float;
+    case DBI_DECIMAL_SIZE8:
+      return (long long)RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_double;
+    default:
+      _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+      return my_ERROR;
+    }
+  case DBI_TYPE_STRING:
+    if (RESULT->rows[RESULT->currowidx]->field_sizes[fieldidx] == 0
+	&& RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_string == NULL) {
+      /* string does not exist */
+      return 0; /* do not raise an error */
+    }
+    /* else if field size == 0: empty string */
+    /* todo: do we need strtoll() error handling? */
+    return strtoll((const char *)(RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_string), NULL, 10);
+  case DBI_TYPE_BINARY:
+    return 0; /* do not raise an error */
+  case DBI_TYPE_DATETIME:
+    return (long long)(RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_datetime);
+  default:
+    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+    return my_ERROR;
+  }
+}
+
+char *dbi_result_get_as_string_copy(dbi_result Result, const char *fieldname) {
+  char *my_ERROR = "ERROR";
+  unsigned int fieldidx;
+  dbi_error_flag errflag = DBI_ERROR_NONE;
+
+  _reset_conn_error(RESULT->conn);
+
+  fieldidx = _find_field(RESULT, fieldname, &errflag);
+  if (errflag != DBI_ERROR_NONE) {
+    dbi_conn_t *conn = RESULT->conn;
+    _error_handler(conn, DBI_ERROR_BADNAME);
+    return strdup(my_ERROR);
+  }
+  return dbi_result_get_as_string_copy_idx(Result, fieldidx+1);
+}
+
+char *dbi_result_get_as_string_copy_idx(dbi_result Result, unsigned int fieldidx) {
+  char *my_ERROR = "ERROR";
+  char *newstring = NULL;
+  char *oldstring = NULL;
+  struct tm utctime;
+
+  fieldidx--;
+
+  newstring = malloc(32); /* sufficient for integers, decimal, and datetime */
+
+  if (!newstring) {
+    _error_handler(RESULT->conn, DBI_ERROR_NOMEM);
+    return my_ERROR;
+  }
+  *newstring = '\0';
+
+  /* output depends on: type, size, sign */
+  switch (RESULT->field_types[fieldidx]) {
+  case DBI_TYPE_INTEGER:
+    switch (RESULT->field_attribs[fieldidx] & DBI_INTEGER_SIZEMASK) {
+    case DBI_INTEGER_SIZE1:
+      if (RESULT->field_attribs[fieldidx] & DBI_INTEGER_UNSIGNED) {
+	snprintf(newstring, 32, "%hu", RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_char);
+      }
+      else {
+	snprintf(newstring, 32, "%hd", RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_char);
+      }
+      break;
+    case DBI_INTEGER_SIZE2:
+      if (RESULT->field_attribs[fieldidx] & DBI_INTEGER_UNSIGNED) {
+	snprintf(newstring, 32, "%hu", RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_short);
+      }
+      else {
+	snprintf(newstring, 32, "%hd", RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_short);
+      }
+      break;
+    case DBI_INTEGER_SIZE3:
+    case DBI_INTEGER_SIZE4:
+      if (RESULT->field_attribs[fieldidx] & DBI_INTEGER_UNSIGNED) {
+	snprintf(newstring, 32, "%u", RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_long);
+      }
+      else {
+	snprintf(newstring, 32, "%d", RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_long);
+      }
+      break;
+    case DBI_INTEGER_SIZE8:
+      if (RESULT->field_attribs[fieldidx] & DBI_INTEGER_UNSIGNED) {
+	snprintf(newstring, 32, "%llu", RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_longlong);
+      }
+      else {
+	snprintf(newstring, 32, "%lld", RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_longlong);
+      }
+      break;
+    default:
+      _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+    }
+    break;
+  case DBI_TYPE_DECIMAL:
+    switch (RESULT->field_attribs[fieldidx] & DBI_DECIMAL_SIZEMASK) {
+    case DBI_DECIMAL_SIZE4:
+      snprintf(newstring, 32, "%e", RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_float);
+      break;
+    case DBI_DECIMAL_SIZE8:
+      snprintf(newstring, 32, "%e", RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_double);
+      break;
+    default:
+      _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+    }
+    break;
+  case DBI_TYPE_STRING:
+    if (RESULT->rows[RESULT->currowidx]->field_sizes[fieldidx] == 0
+	&& RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_string == NULL) {
+      /* string does not exist */
+      /* return an empty string instead, no error */
+    }
+    else {
+      /* else if field size == 0: empty string */
+
+      oldstring = newstring;
+      if ((newstring = strdup(RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_string)) == NULL) {
+	_error_handler(RESULT->conn, DBI_ERROR_NOMEM);
+	newstring = oldstring;
+      }
+      else {
+	/* don't free in case of an error as we'd return a dangling pointer */
+	free(oldstring);
+      }
+    }
+    break;
+  case DBI_TYPE_BINARY:
+    break; /* return empty string, do not raise an error */
+  case DBI_TYPE_DATETIME:
+    gmtime_r(&(RESULT->rows[RESULT->currowidx]->field_values[fieldidx].d_datetime), &utctime);
+    snprintf(newstring, 32, "%04d-%02d-%02d %02d:%02d:%02d", utctime.tm_year+1900, utctime.tm_mon+1, utctime.tm_mday, utctime.tm_hour, utctime.tm_min, utctime.tm_sec);
+    break;
+  default:
+    _error_handler(RESULT->conn, DBI_ERROR_BADTYPE);
+  }
+
+  return newstring; /* is still empty string in case of an error */
+}
+
+
 /* RESULT: bind_* functions */
 
 static int _setup_binding(dbi_result_t *result, const char *fieldname, void *bindto, void *helperfunc) {
@@ -1285,7 +1559,10 @@ static int _setup_binding(dbi_result_t *result, const char *fieldname, void *bin
     _error_handler(/*RESULT->conn*/ NULL, DBI_ERROR_BADPTR);
     return DBI_BIND_ERROR;
   }
-  else if (!fieldname) {
+
+  _reset_conn_error(result->conn);
+
+  if (!fieldname) {
     _error_handler(result->conn, DBI_ERROR_BADNAME);
     return DBI_BIND_ERROR;
   }
@@ -1453,7 +1730,9 @@ static unsigned int _find_field(dbi_result_t *result, const char *fieldname, dbi
 }
 
 static int _is_row_fetched(dbi_result_t *result, unsigned long long row) {
-  if (!result->rows || (row >= result->numrows_matched)) return -1;
+  /* Bull patch reported by Tom Lane */
+  /*  if (!result->rows || (row >= result->numrows_matched)) return -1; */
+  if (!result->rows || (row > result->numrows_matched)) return -1;
   return !(result->rows[row] == NULL);
 }
 
